// stdafx.h : include file for standard system include files,
// or project specific include files that are used frequently, but
// are changed infrequently
//

#pragma once

#include "targetver.h"

#include <windows.h>
#include <stdio.h>
#include <tchar.h>
#include <assert.h>
#include <string>


// Bunch of util function cruft. Everything inline so I can just throw it here in the header.
namespace LeoHelpers
{
	// Handles things like "X:" "X:\" "\\?\X:" "\\?\X:\"
	// Does not handle relative paths, or stuff like "X:\."
	// Remember that "X:" doesn't literally mean the root to many APIs
	// (it means the current dir on drive X, not neccessarily the root),
	// even though it does to most of the code in here.
	inline bool IsDriveRootPath(const wchar_t *szPath, bool bAllowSubDirs, wchar_t *pCharOutDriveLetter)
	{
		if (pCharOutDriveLetter != NULL)
		{
			*pCharOutDriveLetter = L'\0';
		}

		// Skip \\?\ prefix if it's there.
		if (szPath[0] == L'\\'
		&&	szPath[1] == L'\\'
		&&	szPath[2] == L'?'
		&&	szPath[3] == L'\\')
		{
			szPath += 4;
		}

		// Assumption: a-z and A-Z in that order with the character set we're compiled for.
		if ((szPath[0] < L'a' || szPath[0] > L'z')
		&&	(szPath[0] < L'A' || szPath[0] > L'Z'))
		{
			return false;
		}

		if (szPath[1] != L':')
		{
			return false;
		}

		if (szPath[2] == L'\0'
		||	(szPath[2] == L'\\' && (bAllowSubDirs || szPath[3] == L'\0')))
		{
			if (pCharOutDriveLetter != NULL)
			{
				*pCharOutDriveLetter = szPath[0];
			}
			return true;
		}

		return false;
	}

	// If the path is like "X:" turns it into "X:\"
	inline void EnsureSlashAtEndIfDriveRootPath(std::wstring *pstrPath)
	{
		if (pstrPath != NULL
		&&	!pstrPath->empty()
		&&	LeoHelpers::IsDriveRootPath(pstrPath->c_str(), false, NULL)
		&&	pstrPath->at( pstrPath->length() - 1 ) == L':')
		{
			pstrPath->append(L"\\");
		}
	}

	inline void TrimTrailingSlashes(std::wstring &strPath)
	{
		std::wstring::size_type len = strPath.length();

		while(len > 0 && (strPath[len - 1] == L'\\' || strPath[len - 1] == L'/'))
		{
			--len;
			strPath.erase(len);
		}
	}

	inline void GetParentPathString(std::wstring *pstrResult, bool bIncludeSlashIfDriveRoot)
	{
		assert(NULL != pstrResult);

		if (NULL == pstrResult)
		{
			return;
		}

		// Trim trailing slashes so paths like C:\Moo\Cow\ turn into C:\Moo not C:\Moo\Cow.

		LeoHelpers::TrimTrailingSlashes(*pstrResult);

		std::wstring::size_type lastPos = pstrResult->find_last_of(L"\\/");

		if (lastPos == std::wstring::npos)
		{
			pstrResult->clear();
		}
		else
		{
			pstrResult->erase(lastPos);

			LeoHelpers::TrimTrailingSlashes(*pstrResult);
		}

		// Return things like X:\ rather than like X: if asked.
		if (bIncludeSlashIfDriveRoot)
		{
			LeoHelpers::EnsureSlashAtEndIfDriveRootPath(pstrResult);
		}
	}

	inline std::wstring *GetParentPathString(std::wstring *pstrResult, const wchar_t *path, bool bIncludeSlashIfDriveRoot)
	{
		assert(NULL != pstrResult && NULL != path);

		if (NULL == pstrResult || NULL == path)
		{
			return NULL;
		}

		*pstrResult = path;

		GetParentPathString(pstrResult, bIncludeSlashIfDriveRoot);

		return pstrResult;
	}

	// Never pass ArrayScoper objects across process/DLL boundaries.
	// (It's all inline so the new/delete calls may hit different heaps.)
	template <typename T> class ArrayScoper
	{
	protected:
		T *m_p;
	private:
		ArrayScoper(const ArrayScoper &rhs); // disallow
		ArrayScoper &operator=(const ArrayScoper &rhs); // disallow
	public:
		ArrayScoper() : m_p(NULL)        { }
		ArrayScoper(T *p) : m_p(p)       { }
		/* not virtual */ ~ArrayScoper() { Free(); }
		inline bool     IsEmpty() const  { return (m_p == NULL); }
		inline void     Free()           { delete[] m_p; m_p = NULL; }
		inline void     Forget()         { m_p = NULL; }
		inline void     Set(T *p)        { if (p != m_p) { Free(); m_p = p; } }
		inline       T *Get()            { return m_p; }
		inline const T *Get() const      { return m_p; }
		inline       T *GetAndForget()   { T *p = m_p; Forget(); return p; }
	//	operator T*() { return m_p; }
	};

	typedef ArrayScoper< wchar_t > WCharArrayScoper;

	inline bool CreateProcessWrapper(const wchar_t *szExePath, const wchar_t *szArgs, bool bArgsAlreadyHaveExe, const wchar_t *szCurDir, bool bExeDirAsCurDir, HANDLE *phProcess)
	{
		if (phProcess != NULL)
		{
			*phProcess = NULL;
		}

		// Convention requires that the exe name be explicitly included (in quotes) as the first argument.
		// (This is true even when giving CreateProcess a separate exe string.)
		std::wstring strArgs;
		
		if (!bArgsAlreadyHaveExe)
		{
			strArgs = L"\"";
			strArgs += szExePath;
			strArgs += L"\"";
		}

		if (szArgs != NULL && szArgs[0] != L'\0')
		{
			if (!strArgs.empty())
			{
				strArgs += L" ";
			}
			strArgs += szArgs;
		}

		// Unicode CreateProcess is stupid and really needs a non-const argument string as it may stupidly modify it in-place because it is stupid.
		LeoHelpers::WCharArrayScoper scArgsTemp(new wchar_t[strArgs.length() + 1]);
		wcscpy_s(scArgsTemp.Get(), strArgs.length() + 1, strArgs.c_str());

		// If the caller asked to use the target exe's dir as the CD, get it.
		std::wstring strCurDirTemp;
		if (bExeDirAsCurDir)
		{
			assert(szCurDir == NULL);

			LeoHelpers::GetParentPathString(&strCurDirTemp, szExePath, true);
			szCurDir = strCurDirTemp.c_str();

			assert(!strCurDirTemp.empty()); // Makes no sense asking to use the exe's dir then giving an exe with no path.
		}

		STARTUPINFO startupInfo = {0};
		startupInfo.cb = sizeof(startupInfo);
		PROCESS_INFORMATION processInfo = {0};

		// http://stackoverflow.com/questions/4206190/win32-createprocess-when-is-create-unicode-environment-really-needed
		if (!::CreateProcess(szExePath, scArgsTemp.Get(), NULL, NULL, FALSE, CREATE_UNICODE_ENVIRONMENT, NULL, szCurDir, &startupInfo, &processInfo))
		{
			return false;
		}

		::CloseHandle(processInfo.hThread);

		if (phProcess != NULL)
		{
			*phProcess = processInfo.hProcess;
		}
		else
		{
			::CloseHandle(processInfo.hProcess);
		}

		return true;
	}

};
